﻿using System;
using System.Threading;
using Windows.UI.Xaml.Controls;
using Windows.Devices.Enumeration;
using Windows.Devices.I2c;
using System.Diagnostics;
using Windows.UI.Xaml;

namespace RaspberryI2CArduino
{
    public sealed partial class MainPage : Page
    {
        private I2cDevice Device;
        private Timer periodicTimer;

        public MainPage()
        {
            this.InitializeComponent();
            InitI2C();
        }
        private async void InitI2C()
        {
            var settings = new I2cConnectionSettings(0x01); // Indirizzo 1 Arduino
            settings.BusSpeed = I2cBusSpeed.StandardMode;
            string aqs = I2cDevice.GetDeviceSelector("I2C1");
            var dis = await DeviceInformation.FindAllAsync(aqs);
            Device = await I2cDevice.FromIdAsync(dis[0].Id, settings);
            periodicTimer = new Timer(this.TimerCallback, null, 0, 100); // Crea un timer            
        }

        private void TimerCallback(object state)
        {
            byte[] ReadBuf = new byte[5];
            try
            {
                Device.Read(ReadBuf); // Legge i dati
            }
            catch (Exception f)
            {
                Debug.WriteLine(f.Message);
            }
            char[] cArray = System.Text.Encoding.UTF8.GetString(ReadBuf, 0, 5).ToCharArray();  // Converte Byte a Char
            String c = new String(cArray);
            Debug.WriteLine(c);

            var task = this.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                TextTemp.Text = c;
            });
        }

        
        private void toggleButton_Checked_1(object sender, Windows.UI.Xaml.RoutedEventArgs e)
        {
            byte[] WriteBuf = new byte[1];
            WriteBuf[0] = 1;
            Device.Write(WriteBuf);
            toggleButton.Content = "LED ON";
        }

        private void toggleButton_Unchecked_1(object sender, Windows.UI.Xaml.RoutedEventArgs e)
        {
            byte[] WriteBuf = new byte[1];
            WriteBuf[0] = 0;
            Device.Write(WriteBuf);
            toggleButton.Content = "LED OFF";
        }

        private void button_Click(object sender, Windows.UI.Xaml.RoutedEventArgs e)
        {
            Application.Current.Exit();
        }
    }
}
