// RPi-PubNub by Pier Calderan - 2015

using Newtonsoft.Json.Linq;
using PubNubMessaging.Core;
using System;
using System.Diagnostics;
using Windows.Data.Json;
using Windows.Devices.Enumeration;
using Windows.Devices.Gpio;
using Windows.Devices.I2c;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.Devices.Spi;

namespace RPiPubNub
{
    public sealed partial class MainPage : Page
    {
        private class PubNubMessage
        {
        }
        // Timer
        private DispatcherTimer TimerRoutine;
        // CAMBIARE LE CHIAVI PubNub 
        Pubnub pubnub = new Pubnub("pub-c-f280f7a4-df81-464f-be54-09a240eaa422", "sub-c-bc983112-2ef1-11e5-ae01-02ee2ddab7fe");
        Random rnd = new Random();

        public MainPage()
        {
            this.InitializeComponent();
            SubscribePubNubMessages();
            InitSPI();
            // Timer ogni 3 seconds
            TimerRoutine = new DispatcherTimer();
            TimerRoutine.Interval = TimeSpan.FromMilliseconds(3000);
            TimerRoutine.Tick += Timer_Tick;
            TimerRoutine.Start();
        }
        
        private async void InitSPI()
        {
            try
            {
                var settings = new SpiConnectionSettings(SPI_CHIP_SELECT_LINE);
                settings.ClockFrequency = 500000;
                settings.Mode = SpiMode.Mode0;
                string spiAqs = SpiDevice.GetDeviceSelector(SPI_CONTROLLER_NAME);
                var deviceInfo = await DeviceInformation.FindAllAsync(spiAqs);
                SpiDisplay = await SpiDevice.FromIdAsync(deviceInfo[0].Id, settings);
            }

            catch (Exception ex)
            {
                throw new Exception("SPI Initialization Failed", ex);
            }
        }
       
        public void DisplayTextBoxContents()
        {
            SpiDisplay.TransferFullDuplex(writeBuffer, readBuffer);
            res = ConvertToCelsius(readBuffer);
            gradi.Text = res.ToString() + " C";
        }

        public int ConvertToCelsius(byte[] data)
        {
            int result = data[1] & 0x0F;
            result <<= 8;
            result += data[2];
            result /= 4;

            double step = 3.2 / 1024;
            double temperatura = result * step * 100;
            result = Convert.ToInt32(temperatura);
            Debug.WriteLine(result);
            return result;
        }

        private const string SPI_CONTROLLER_NAME = "SPI0";  /* Raspberry Pi 2, SPI0                             */
        private const Int32 SPI_CHIP_SELECT_LINE = 0;
        byte[] readBuffer = new byte[3];
        byte[] writeBuffer = new byte[3] { 0x06, 0x00, 0x00 };
        private SpiDevice SpiDisplay;
        int res;

        private void SubscribePubNubMessages()
        {
            pubnub.Subscribe<string>("rpipb-control", PubNubSubscribeSuccess, DisplaySubscribeConnectStatusMessage, PubNubError);
        }

        // Timer routine to read sensor value and send pubnub message
        private void Timer_Tick(object sender, object e)
        {
            DisplayTextBoxContents();
            JsonObject json = new JsonObject();
            int temperature = ConvertToCelsius (readBuffer);
            json["Temperature"] = JsonValue.CreateNumberValue(temperature);
            String jsonString = json.Stringify();
            pubnub.Publish<string>("raspi-temperatura", jsonString, PubNubSucess, PubNubError); // il nome del canale
        }

        // Receive PubNub messages
        private static void PubNubSubscribeSuccess(string publishResult)
        {
            Debug.WriteLine("Message: " + publishResult);
            JArray message = JArray.Parse(publishResult);
            Debug.WriteLine("Pin: " + message[0]["pin"] + ", Value: " + message[0]["value"]);
        }

        private static void DisplaySubscribeConnectStatusMessage(string publishResult)
        {
            Debug.WriteLine("Connection: " + publishResult);
        }
        // Successfully send the message, simply display it on the console
        private static void PubNubSucess(string publishResult)
        {
            Debug.WriteLine("Success: " + publishResult);
        }

        // We have some issue sending the message, simply display it on the console
        private static void PubNubError(PubnubClientError error)
        {
            Debug.WriteLine("Error: " + error.ToString());
        }
    }
}